/*
 * Copyright (c) 2024 DTB <trinity@trinity.moe>
 * SPDX-License-Identifier: AGPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/.
 */

#include <errno.h> /* errno */
#include <stdio.h> /* fclose(3), feof(3), fgetc(3), fgets(3), fopen(3),
                    * fprintf(3), fputc(3), stderr, stdin, stdout, EOF, FILE,
                    * NULL */
#include <stdlib.h> /* size_t */
#include <string.h> /* strchr(3), strcmp(3), strerror(3), strtok(3) */

#if !defined EX_OK || !defined EX_SOFTWARE || !defined EX_USAGE 
#	include <sysexits.h>
#endif

/* Done while looking at plan9ports' src/cmd/p.c.
 * Taken from plan9 are the interface (minus p(1)'s shell escaping and the
 * ability to page multiple files) and the practice of reading /dev/tty to get
 * keyboard input without going through stdin. Also, a newline is now printed
 * before querying the keyboard. */

static char *whitespace = " \n\r\t\v";

static struct {
	size_t quantity;
	enum { LINES = 0, BYTES = 1 } type;
} default_page_unit = { 22, LINES } /* Plan 9 default */;
static FILE *input;
static char *prompt = ": ";

static char *program_name = "pg";

/* Commands start with cmd_. They take an argc and NULL-terminated argv, like
 * main, and return one of the following RET_ enums as a status. */
enum{ RET_ARGSERROR, RET_OK, RET_NOCMD, RET_REQSHUTDOWN, RET_SOFTWARE };

static int page_bytes(FILE *f, size_t l){
	int c;

	while((c = fgetc(f)) != EOF)
		if((c = fputc(c, stdout)) == EOF || --l == 0)
			break;

	return c == EOF
	    ? RET_REQSHUTDOWN
	    : RET_OK;
}

static int page_lines(FILE *f, size_t l){
	int c;

	while((c = fgetc(f)) != EOF)
		if((c = fputc(c, stdout)) == EOF || (l -= (c == '\n')) == 0)
			break;

	return c == EOF
	    ? RET_REQSHUTDOWN
	    : RET_OK;
}

static int cmd_quit(int argc, char **argv){ return RET_REQSHUTDOWN; }
static int cmd_default_page(int argc, char **argv){

	if(strtok(NULL, whitespace) != NULL)
		return RET_ARGSERROR; /* This shouldn't be possible. */
	if(default_page_unit.type == BYTES)
		return page_bytes(input, default_page_unit.quantity);
	else if(default_page_unit.type == LINES)
		return page_lines(input, default_page_unit.quantity);
	else
		return RET_SOFTWARE;
}

/* A CmdMap must be NULL-terminated. */
static struct CmdMap{ char *name; int (*fn)(int, char **); } cmd_map[] = {
	{ "", cmd_default_page },

	/* don't make the user feel trapped */
	{ "exit", cmd_quit }, { "q", cmd_quit }, { ":q", cmd_quit },
	{ ":q!", cmd_quit }, { "quit", cmd_quit }, { "ZZ", cmd_quit },
	
	{ NULL, NULL }
};

#define ARGV_MAX 10
/* Find and execute the command in the command map, given a corresponding
 * command line. */
static int cmdline_exec(struct CmdMap *map, char *cmdline){
	static int argc;
	static char *argv[ARGV_MAX];

	if((argv[(argc = 0)] = strtok(cmdline, whitespace)) == NULL){
		while(cmdline[0] != '\0')
			cmdline = &cmdline[1];
		argv[argc] = cmdline;
		argv[++argc] = NULL;
	}else{
		while((argv[++argc] = strtok(NULL , whitespace)) != NULL
		      && argc < ARGV_MAX);
	}

	for(; map->name != NULL; map = &map[1])
		if(strcmp(map->name, argv[0]) == 0)
			return map->fn(argc, argv);

	fprintf(stderr, "%s: %s: not found\n", program_name, argv[0]);
	return RET_NOCMD;
}

int main(int argc, char *argv[]){
	unsigned char cmd[99+1];
	int r;
	FILE *t;

	input = stdin;

	if(argc > 0)
		program_name = argv[0];

	if(argc > 1){
		fprintf(stderr, "Usage: %s\n", program_name);
		return EX_USAGE;
	}

	if((t = fopen("/dev/tty", "rb")) == NULL){
		fprintf(stderr, "%s: /dev/tty: %s\n", program_name, strerror(errno));
		return EX_SOFTWARE;
	}

	r = 0;
	for(;;){
		if(r != RET_REQSHUTDOWN){
			fputs(prompt, stderr);
			fgets((char *)cmd, (sizeof cmd) / (sizeof *cmd), t);
			if(strchr((char *)cmd, '\n') == NULL){ /* fast-forward stream */
				int c;

				while((c = fgetc(t)) != '\n')
					if(c == EOF)
						break;
			}
		}
		if(r == RET_REQSHUTDOWN || feof(t)){
			fclose(t);
			return EX_OK;
		}

		switch((r = cmdline_exec(cmd_map, (char *)cmd))){
		case RET_ARGSERROR: case RET_NOCMD: case RET_OK:
			break;
		case RET_REQSHUTDOWN: case RET_SOFTWARE: default:
			r = RET_REQSHUTDOWN;
		}
	}

	/* UNREACHABLE */
	return EX_SOFTWARE;
}
